@@ -1,5 +1,5 @@
 {-# LANGUAGE DeriveFunctor #-}
-{-# LANGUAGE DeriveTraversable  #-}
+{-# LANGUAGE DeriveTraversable #-}
 {-# LANGUAGE DeriveFoldable #-}
 {-# LANGUAGE DeriveDataTypeable #-}
 {-# LANGUAGE DeriveGeneric #-}
@@ -12,14 +12,14 @@
 -- support for things like JSON request and response bodies. For most users,
 -- this will be an easier place to start. You can read the tutorial at:
 --
--- https://github.com/commercialhaskell/jump/blob/master/doc/http-client.md
+-- https://haskell-lang.org/library/http-client
 --
 -- = Lower-level API
 --
 -- This is the main entry point for using http-client. Used by itself, this
 -- module provides low-level access for streaming request and response bodies,
 -- and only non-secure HTTP connections. Helper packages such as http-conduit
--- provided higher level streaming approaches, while other helper packages like
+-- provide higher level streaming approaches, while other helper packages like
 -- http-client-tls provide secure connections.
 --
 -- There are three core components to be understood here: requests, responses,
@@ -67,7 +67,7 @@
 -- be assumed to throw an 'HttpException' in the event of some problem, and all
 -- pure functions will be total. For example, 'withResponse', 'httpLbs', and
 -- 'BodyReader' can all throw exceptions. Functions like 'responseStatus' and
--- 'applyBasicAuth' are guaranteed to be total (or there\'s a bug in the
+-- 'applyBasicAuth' are guaranteed to be total (or there's a bug in the
 -- library).
 --
 -- One thing to be cautioned about: the type of 'parseRequest' allows it to work in
@@ -75,12 +75,6 @@
 -- the case of an invalid URI. In addition, if you leverage the @IsString@
 -- instance of the 'Request' value via @OverloadedStrings@, an invalid URI will
 -- result in a partial value. Caveat emptor!
---
--- Non-2xx responses: the default behavior of all functions in http-client is
--- to automatically perform up to 10 redirects (response codes 301, 302, 303,
--- and 307), and to throw a 'StatusCodeException' on all responses whose status
--- are not in the 2xx range. These behaviors can be overridden by the
--- 'redirectCount' and 'checkStatus' settings on a request, respectively.
 module Network.HTTP.Client
     ( -- $example1
 
@@ -111,9 +105,10 @@ module Network.HTTP.Client
     , managerTlsConnection
     , managerResponseTimeout
     , managerRetryableException
-    , managerWrapIOException
+    , managerWrapException
     , managerIdleConnectionCount
     , managerModifyRequest
+    , managerModifyResponse
       -- *** Manager proxy settings
     , managerSetProxy
     , managerSetInsecureProxy
@@ -125,9 +120,16 @@ module Network.HTTP.Client
     , proxyEnvironment
     , proxyEnvironmentNamed
     , defaultProxy
+      -- *** Response timeouts
+    , ResponseTimeout
+    , responseTimeoutMicro
+    , responseTimeoutNone
+    , responseTimeoutDefault
       -- *** Helpers
     , rawConnectionModifySocket
+    , rawConnectionModifySocketSize
       -- * Request
+      -- $parsing-request
     , parseUrl
     , parseUrlThrow
     , parseRequest
@@ -153,7 +155,7 @@ module Network.HTTP.Client
     , applyBasicProxyAuth
     , decompress
     , redirectCount
-    , checkStatus
+    , checkResponse
     , responseTimeout
     , cookieJar
     , requestVersion
@@ -177,8 +179,12 @@ module Network.HTTP.Client
     , brRead
     , brReadSome
     , brConsume
+      -- * Advanced connection creation
+    , makeConnection
+    , socketConnection
       -- * Misc
     , HttpException (..)
+    , HttpExceptionContent (..)
     , Cookie (..)
     , CookieJar
     , Proxy (..)
@@ -187,6 +193,7 @@ module Network.HTTP.Client
     ) where
 
 import Network.HTTP.Client.Body
+import Network.HTTP.Client.Connection (makeConnection, socketConnection)
 import Network.HTTP.Client.Cookies
 import Network.HTTP.Client.Core
 import Network.HTTP.Client.Manager
@@ -194,7 +201,6 @@ import Network.HTTP.Client.Request
 import Network.HTTP.Client.Response
 import Network.HTTP.Client.Types
 
-import Data.Text (Text)
 import Data.IORef (newIORef, writeIORef, readIORef, modifyIORef)
 import qualified Data.ByteString.Lazy as L
 import Data.Foldable (Foldable)
@@ -202,7 +208,7 @@ import Data.Traversable (Traversable)
 import Network.HTTP.Types (statusCode)
 import GHC.Generics (Generic)
 import Data.Typeable (Typeable)
-import Control.Exception (bracket)
+import Control.Exception (bracket, handle, throwIO)
 
 -- | A datatype holding information on redirected requests and the final response.
 --
@@ -222,7 +228,7 @@ data HistoriedResponse body = HistoriedResponse
     --
     -- Since 0.4.1
     }
-    deriving (Functor, Traversable, Foldable, Show, Typeable, Generic)
+    deriving (Functor, Data.Traversable.Traversable, Data.Foldable.Foldable, Show, Typeable, Generic)
 
 -- | A variant of @responseOpen@ which keeps a history of all redirects
 -- performed in the interim, together with the first 1024 bytes of their
@@ -230,12 +236,16 @@ data HistoriedResponse body = HistoriedResponse
 --
 -- Since 0.4.1
 responseOpenHistory :: Request -> Manager -> IO (HistoriedResponse BodyReader)
-responseOpenHistory req0 man = do
+responseOpenHistory req0 man0 = handle (throwIO . toHttpException req0) $ do
     reqRef <- newIORef req0
     historyRef <- newIORef id
     let go req0 = do
-            (man, req) <- getModifiedRequestManager man req0
-            (req', res) <- httpRaw' req man
+            (man, req) <- getModifiedRequestManager man0 req0
+            (req', res') <- httpRaw' req man
+            let res = res'
+                    { responseBody = handle (throwIO . toHttpException req0)
+                                            (responseBody res')
+                    }
             case getRedirectedRequest
                     req'
                     (responseHeaders res)
@@ -288,8 +298,6 @@ managerSetSecureProxy po m = m { managerProxySecure = po }
 managerSetProxy :: ProxyOverride -> ManagerSettings -> ManagerSettings
 managerSetProxy po = managerSetInsecureProxy po . managerSetSecureProxy po
 
-
-
 -- $example1
 -- = Example Usage
 --
@@ -329,3 +337,33 @@ managerSetProxy po = managerSetInsecureProxy po . managerSetSecureProxy po
 -- >   putStrLn $ "The status code was: " ++ (show $ statusCode $ responseStatus response)
 -- >   print $ responseBody response
 --
+
+-- | Specify a response timeout in microseconds
+--
+-- @since 0.5.0
+responseTimeoutMicro :: Int -> ResponseTimeout
+responseTimeoutMicro = ResponseTimeoutMicro
+
+-- | Do not have a response timeout
+--
+-- @since 0.5.0
+responseTimeoutNone :: ResponseTimeout
+responseTimeoutNone = ResponseTimeoutNone
+
+-- | Use the default response timeout
+--
+-- When used on a 'Request', means: use the manager's timeout value
+--
+-- When used on a 'ManagerSettings', means: default to 30 seconds
+--
+-- @since 0.5.0
+responseTimeoutDefault :: ResponseTimeout
+responseTimeoutDefault = ResponseTimeoutDefault
+
+-- $parsing-request
+--
+-- The way you parse string of characters to construct a 'Request' will
+-- determine whether exceptions will be thrown on non-2XX response status
+-- codes. This is because the behavior is controlled by a setting in
+-- 'Request' itself (see 'checkResponse') and different parsing functions
+-- set it to different 'IO' actions.
