@@ -10,7 +10,6 @@ module Network.HTTP.Client.Core
     , getModifiedRequestManager
     , responseOpen
     , responseClose
-    , applyCheckStatus
     , httpRedirect
     , httpRedirect'
     ) where
@@ -18,21 +17,20 @@ module Network.HTTP.Client.Core
 #if !MIN_VERSION_base(4,6,0)
 import Prelude hiding (catch)
 #endif
-import Network.HTTP.Client.Body
-import Network.HTTP.Client.Cookies
+import Network.HTTP.Types
 import Network.HTTP.Client.Manager
+import Network.HTTP.Client.Types
+import Network.HTTP.Client.Body
 import Network.HTTP.Client.Request
 import Network.HTTP.Client.Response
-import Network.HTTP.Client.Types
-import Network.HTTP.Types
+import Network.HTTP.Client.Cookies
 import Data.Maybe (fromMaybe, isJust)
 import Data.Time
 import Control.Exception
-import qualified Data.ByteString as S
-import qualified Data.ByteString.Char8 as S8
 import qualified Data.ByteString.Lazy as L
 import Data.Monoid
 import Control.Monad (void)
+import System.Timeout (timeout)
 
 -- | Perform a @Request@ using a connection acquired from the given @Manager@,
 -- and then provide the @Response@ to the given function. This function is
@@ -94,11 +92,9 @@ httpRaw' req0 m = do
         Just cj -> do
             now <- getCurrentTime
             return $ insertCookiesIntoRequest req' (evictExpiredCookies cj now) now
-        Nothing -> return (req', mempty)
+        Nothing -> return (req', Data.Monoid.mempty)
     (timeout', (connRelease, ci, isManaged)) <- getConnectionWrapper
-        req
         (responseTimeout' req)
-        (failedConnectionException req)
         (getConn req m)
 
     -- Originally, we would only test for exceptions when sending the request,
@@ -126,12 +122,31 @@ httpRaw' req0 m = do
                 return (req, res {responseCookieJar = cookie_jar})
             Nothing -> return (req, res)
   where
+    getConnectionWrapper mtimeout f =
+        case mtimeout of
+            Nothing -> fmap ((,) Nothing) f
+            Just timeout' -> do
+                before <- getCurrentTime
+                mres <- timeout timeout' f
+                case mres of
+                    Nothing -> throwHttp ConnectionTimeout
+                    Just res -> do
+                        now <- getCurrentTime
+                        let timeSpentMicro = diffUTCTime now before * 1000000
+                            remainingTime = round $ fromIntegral timeout' - timeSpentMicro
+                        if remainingTime <= 0
+                            then throwHttp ConnectionTimeout
+                            else return (Just remainingTime, res)
 
-    responseTimeout' req
-        | rt == useDefaultTimeout = mResponseTimeout m
-        | otherwise = rt
-      where
-        rt = responseTimeout req
+    responseTimeout' req =
+        case responseTimeout req of
+            ResponseTimeoutDefault ->
+                case mResponseTimeout m of
+                    ResponseTimeoutDefault -> Just 30000000
+                    ResponseTimeoutNone -> Nothing
+                    ResponseTimeoutMicro u -> Just u
+            ResponseTimeoutNone -> Nothing
+            ResponseTimeoutMicro u -> Just u
 
 -- | The used Manager can be overridden (by requestManagerOverride) and the used
 -- Request can be modified (through managerModifyRequest). This function allows
@@ -176,22 +191,22 @@ getModifiedRequestManager manager0 req0 = do
 --
 -- Since 0.1.0
 responseOpen :: Request -> Manager -> IO (Response BodyReader)
-responseOpen req0 manager' = handle addTlsHostPort $ mWrapIOException manager $ do
-    (req, res) <-
-        if redirectCount req0 == 0
-            then httpRaw' req0 manager
-            else go (redirectCount req0) req0
-    maybe (return res) throwIO =<< applyCheckStatus req (checkStatus req) res
+responseOpen inputReq manager' = do
+  (manager, req0) <- getModifiedRequestManager manager' inputReq
+  wrapExc req0 $ mWrapException manager req0 $ do
+    (req, res) <- go manager (redirectCount req0) req0
+    checkResponse req req res
+    mModifyResponse manager res
+        { responseBody = wrapExc req0 (responseBody res)
+        }
   where
-    manager = fromMaybe manager' (requestManagerOverride req0)
+    wrapExc :: Request -> IO a -> IO a
+    wrapExc req0 = handle $ throwIO . toHttpException req0
 
-    addTlsHostPort (TlsException e) = throwIO $ TlsExceptionHostPort e (host req0) (port req0)
-    addTlsHostPort e = throwIO e
-
-    go count req' = httpRedirect'
+    go manager0 count req' = httpRedirect'
       count
       (\req -> do
-        (manager, modReq) <- getModifiedRequestManager manager req
+        (manager, modReq) <- getModifiedRequestManager manager0 req
         (req'', res) <- httpRaw' modReq manager
         let mreq = if redirectCount modReq == 0
               then Nothing
@@ -199,41 +214,6 @@ responseOpen req0 manager' = handle addTlsHostPort $ mWrapIOException manager $
         return (res, fromMaybe req'' mreq, isJust mreq))
       req'
 
--- | Apply 'Request'\'s 'checkStatus' and return resulting exception if any.
-applyCheckStatus
-    :: Request
-    -> (Status -> ResponseHeaders -> CookieJar -> Maybe SomeException)
-    -> Response BodyReader
-    -> IO (Maybe SomeException)
-applyCheckStatus req checkStatus' res =
-    case checkStatus' (responseStatus res) (responseHeaders res) (responseCookieJar res) of
-        Nothing -> return Nothing
-        Just exc -> do
-            exc' <-
-                case fromException exc of
-                    Just (StatusCodeException s hdrs cookie_jar) -> do
-                        lbs <- brReadSome (responseBody res) 1024
-                        return $ toException $ StatusCodeException s (hdrs ++
-                            [ ("X-Response-Body-Start", toStrict' lbs)
-                            , ("X-Request-URL", S.concat
-                                [ method req
-                                , " "
-                                , S8.pack $ show $ getUri req
-                                ])
-                            ]) cookie_jar
-                    _ -> return exc
-            responseClose res
-            return (Just exc')
-  where
-#ifndef MIN_VERSION_bytestring
-#define MIN_VERSION_bytestring(x,y,z) 1
-#endif
-#if MIN_VERSION_bytestring(0,10,0)
-    toStrict' = L.toStrict
-#else
-    toStrict' = S.concat . L.toChunks
-#endif
-
 -- | Redirect loop.
 httpRedirect
      :: Int -- ^ 'redirectCount'
@@ -257,7 +237,7 @@ httpRedirect'
      -> IO (Request, Response BodyReader)
 httpRedirect' count0 http' req0 = go count0 req0 []
   where
-    go count _ ress | count < 0 = throwIO $ TooManyRedirects ress
+    go count _ ress | count < 0 = throwHttp $ TooManyRedirects ress
     go count req' ress = do
         (res, req, isRedirect) <- http' req'
         if isRedirect then do
@@ -270,7 +250,15 @@ httpRedirect' count0 http' req0 = go count0 req0 []
                 -- The connection may already be closed, e.g.
                 -- when using withResponseHistory. See
                 -- https://github.com/snoyberg/http-client/issues/169
-                `catch` \(_ :: ConnectionClosed) -> return L.empty
+                `catch` \se ->
+                    case () of
+                      ()
+                        | Just ConnectionClosed <-
+                            fmap unHttpExceptionContentWrapper
+                            (fromException se) -> return L.empty
+                        | Just (HttpExceptionRequest _ ConnectionClosed) <-
+                            fromException se -> return L.empty
+                      _ -> throwIO se
             responseClose res
 
             -- And now perform the actual redirect
