@@ -10,9 +10,9 @@ module Network.HTTP.Client.Manager
     , closeManager
     , withManager
     , getConn
-    , failedConnectionException
     , defaultManagerSettings
     , rawConnectionModifySocket
+    , rawConnectionModifySocketSize
     , proxyFromRequest
     , noProxy
     , useProxy
@@ -30,38 +30,30 @@ import Prelude hiding (catch)
 #endif
 import Control.Applicative ((<|>))
 import Control.Arrow (first)
-import Data.Monoid (mappend)
-import System.IO (hClose, hFlush, IOMode(..))
 import qualified Data.IORef as I
 import qualified Data.Map as Map
 
 import qualified Data.ByteString.Char8 as S8
-import qualified Data.ByteString.Lazy as L
-
-import qualified Blaze.ByteString.Builder as Blaze
 
 import Data.Char (toLower)
 import Data.Text (Text)
 import qualified Data.Text as T
 import Data.Text.Read (decimal)
 
-import Control.Monad.IO.Class (MonadIO, liftIO)
-import Control.Monad (unless, join, when, void, mplus)
-import Control.Exception (mask_, SomeException, bracket, catch, throwIO, fromException, mask, IOException, Exception (..), handle)
+import Control.Monad (unless, join, void)
+import Control.Exception (mask_, catch, throwIO, fromException, mask, IOException, Exception (..), handle)
 import Control.Concurrent (forkIO, threadDelay)
-import Data.Time (UTCTime (..), Day (..), DiffTime, getCurrentTime, addUTCTime)
-import Control.DeepSeq (deepseq)
+import Data.Time (UTCTime (..), getCurrentTime, addUTCTime)
 
 import qualified Network.Socket as NS
 
-import Data.Maybe (mapMaybe)
-import System.IO (Handle)
 import System.Mem.Weak (Weak, deRefWeak)
 import Network.HTTP.Types (status200)
 import Network.HTTP.Client.Types
 import Network.HTTP.Client.Connection
 import Network.HTTP.Client.Headers (parseStatusHeaders)
 import Network.HTTP.Client.Request (applyBasicProxyAuth, extractBasicAuthInfo)
+import Network.HTTP.Proxy
 import Control.Concurrent.MVar (MVar, takeMVar, tryPutMVar, newEmptyMVar)
 import System.Environment (getEnvironment)
 import qualified Network.URI as U
@@ -78,11 +70,12 @@ rawConnectionModifySocket = return . openSocketConnection
 
 -- | Same as @rawConnectionModifySocket@, but also takes in a chunk size.
 --
--- Since 0.4.5
+-- @since 0.5.2
 rawConnectionModifySocketSize :: (NS.Socket -> IO ())
                               -> IO (Int -> Maybe NS.HostAddress -> String -> Int -> IO Connection)
 rawConnectionModifySocketSize = return . openSocketConnectionSize
 
+
 -- | Default value for @ManagerSettings@.
 --
 -- Note that this value does /not/ have support for SSL/TLS. If you need to
@@ -94,14 +87,14 @@ defaultManagerSettings :: ManagerSettings
 defaultManagerSettings = ManagerSettings
     { managerConnCount = 10
     , managerRawConnection = return $ openSocketConnection (const $ return ())
-    , managerTlsConnection = return $ \_ _ _ -> throwIO TlsNotSupported
-    , managerTlsProxyConnection = return $ \_ _ _ _ _ _ -> throwIO TlsNotSupported
-    , managerResponseTimeout = Just 30000000
+    , managerTlsConnection = return $ \_ _ _ -> throwHttp TlsNotSupported
+    , managerTlsProxyConnection = return $ \_ _ _ _ _ _ -> throwHttp TlsNotSupported
+    , managerResponseTimeout = ResponseTimeoutDefault
     , managerRetryableException = \e ->
         case fromException e of
             Just (_ :: IOException) -> True
             _ ->
-                case fromException e of
+                case fmap unHttpExceptionContentWrapper $ fromException e of
                     -- Note: Some servers will timeout connections by accepting
                     -- the incoming packets for the new request, but closing
                     -- the connection as soon as we try to read. To make sure
@@ -110,14 +103,15 @@ defaultManagerSettings = ManagerSettings
                     Just NoResponseDataReceived -> True
                     Just IncompleteHeaders -> True
                     _ -> False
-    , managerWrapIOException =
+    , managerWrapException = \_req ->
         let wrapper se =
                 case fromException se of
-                    Just e -> toException $ InternalIOException e
-                    Nothing -> se
-         in handle $ throwIO . wrapper
+                    Just (_ :: IOException) -> throwHttp $ InternalException se
+                    Nothing -> throwIO se
+         in handle wrapper
     , managerIdleConnectionCount = 512
     , managerModifyRequest = return
+    , managerModifyResponse = return
     , managerProxyInsecure = defaultProxy
     , managerProxySecure = defaultProxy
     }
@@ -199,9 +193,10 @@ newManager ms = do
             , mTlsConnection = tlsConnection
             , mTlsProxyConnection = tlsProxyConnection
             , mRetryableException = managerRetryableException ms
-            , mWrapIOException = managerWrapIOException ms
+            , mWrapException = managerWrapException ms
             , mIdleConnectionCount = managerIdleConnectionCount ms
             , mModifyRequest = managerModifyRequest ms
+            , mModifyResponse = managerModifyResponse ms
             , mSetProxy = \req ->
                 if secure req
                     then httpsProxy req
@@ -252,6 +247,7 @@ reap baton wmapRef =
                     Nothing -> keep
                     Just x -> keep . ((connkey, x):)
 
+    {- FIXME why isn't this being used anymore?
     flushStaleCerts now =
         Map.fromList . mapMaybe flushStaleCerts' . Map.toList
       where
@@ -283,6 +279,7 @@ reap baton wmapRef =
 
         seqDT :: DiffTime -> b -> b
         seqDT = seq
+    -}
 
 neToList :: NonEmptyList a -> [(UTCTime, a)]
 neToList (One a t) = [(t, a)]
@@ -318,7 +315,7 @@ closeManager' connsRef = mask_ $ do
     !m <- I.atomicModifyIORef connsRef $ \x -> (ManagerClosed, x)
     case m of
         ManagerClosed -> return ()
-        ManagerOpen _ m -> mapM_ (nonEmptyMapM_ safeConnClose) $ Map.elems m
+        ManagerOpen _ m' -> mapM_ (nonEmptyMapM_ safeConnClose) $ Map.elems m'
 
 -- | Create, use and close a 'Manager'.
 --
@@ -369,24 +366,16 @@ getManagedConn man key open = mask $ \restore -> do
             I.writeIORef toReuseRef r
             releaseHelper
 
-        releaseHelper = mask $ \restore -> do
+        releaseHelper = mask $ \restore' -> do
             wasReleased <- I.atomicModifyIORef wasReleasedRef $ \x -> (True, x)
             unless wasReleased $ do
                 toReuse <- I.readIORef toReuseRef
-                restore $ case toReuse of
+                restore' $ case toReuse of
                     Reuse -> putSocket man key ci
                     DontReuse -> connectionClose ci
 
     return (connRelease, ci, isManaged)
 
--- | Create an exception to be thrown if the connection for the given request
--- fails.
-failedConnectionException :: Request -> HttpException
-failedConnectionException req =
-    FailedConnectionException host' port'
-  where
-    (_, host', port') = getConnDest req
-
 getConnDest :: Request -> (Bool, String, Int)
 getConnDest req =
     case proxy req of
@@ -397,13 +386,13 @@ getConnDest req =
 -- secure proxy.
 dropProxyAuthSecure :: Request -> Request
 dropProxyAuthSecure req
-    | secure req && useProxy = req
+    | secure req && useProxy' = req
         { requestHeaders = filter (\(k, _) -> k /= "Proxy-Authorization")
                                   (requestHeaders req)
         }
     | otherwise = req
   where
-    (useProxy, _, _) = getConnDest req
+    (useProxy', _, _) = getConnDest req
 
 getConn :: Request
         -> Manager
@@ -411,30 +400,29 @@ getConn :: Request
 getConn req m
     -- Stop Mac OS X from getting high:
     -- https://github.com/snoyberg/http-client/issues/40#issuecomment-39117909
-    | S8.null h = throwIO $ InvalidDestinationHost h
+    | S8.null h = throwHttp $ InvalidDestinationHost h
     | otherwise =
         getManagedConn m (ConnKey connKeyHost connport (host req) (port req) (secure req)) $
             wrapConnectExc $ go connaddr connhost connport
   where
     h = host req
-    (useProxy, connhost, connport) = getConnDest req
+    (useProxy', connhost, connport) = getConnDest req
     (connaddr, connKeyHost) =
-        case (hostAddress req, useProxy) of
+        case (hostAddress req, useProxy') of
             (Just ha, False) -> (Just ha, HostAddress ha)
             _ -> (Nothing, HostName $ T.pack connhost)
 
     wrapConnectExc = handle $ \e ->
-        throwIO $ FailedConnectionException2 connhost connport (secure req)
-            (toException (e :: IOException))
+        throwHttp $ ConnectionFailure (toException (e :: IOException))
     go =
-        case (secure req, useProxy) of
+        case (secure req, useProxy') of
             (False, _) -> mRawConnection m
             (True, False) -> mTlsConnection m
             (True, True) ->
                 let ultHost = host req
                     ultPort = port req
-                    proxyAuthorizationHeader = maybe "" (\h -> S8.concat ["Proxy-Authorization: ", h, "\r\n"]) . lookup "Proxy-Authorization" $ requestHeaders req
-                    hostHeader = S8.concat ["Host: ", ultHost, (S8.pack $ show ultPort), "\r\n"]
+                    proxyAuthorizationHeader = maybe "" (\h' -> S8.concat ["Proxy-Authorization: ", h', "\r\n"]) . lookup "Proxy-Authorization" $ requestHeaders req
+                    hostHeader = S8.concat ["Host: ", ultHost, ":", (S8.pack $ show ultPort), "\r\n"]
                     connstr = S8.concat
                         [ "CONNECT "
                         , ultHost
@@ -446,9 +434,9 @@ getConn req m
                         , "\r\n"
                         ]
                     parse conn = do
-                        sh@(StatusHeaders status _ _) <- parseStatusHeaders conn Nothing Nothing
+                        StatusHeaders status _ _ <- parseStatusHeaders conn Nothing Nothing
                         unless (status == status200) $
-                            throwIO $ ProxyConnectException ultHost ultPort $ Right $ StatusCodeException status [] (CJ [])
+                            throwHttp $ ProxyConnectException ultHost ultPort status
                  in mTlsProxyConnection m connstr parse (S8.unpack ultHost)
 
 -- | Get the proxy settings from the @Request@ itself.
@@ -477,13 +465,8 @@ useProxy p = ProxyOverride $ const $ return $ \req -> req { proxy = Just p }
 -- Since 0.4.7
 proxyEnvironment :: Maybe Proxy -- ^ fallback if no environment set
                  -> ProxyOverride
-proxyEnvironment mp = ProxyOverride $ \secure ->
-    envHelper (envName secure) $ maybe EHNoProxy EHUseProxy mp
-
-envName :: Bool -- ^ secure?
-        -> Text
-envName False = "http_proxy"
-envName True = "https_proxy"
+proxyEnvironment mp = ProxyOverride $ \secure' ->
+    systemProxyHelper Nothing (httpProtocol secure') $ maybe EHNoProxy EHUseProxy mp
 
 -- | Same as 'proxyEnvironment', but instead of default environment variable
 -- names, allows you to set your own name.
@@ -493,65 +476,12 @@ proxyEnvironmentNamed
     :: Text -- ^ environment variable name
     -> Maybe Proxy -- ^ fallback if no environment set
     -> ProxyOverride
-proxyEnvironmentNamed name =
-    ProxyOverride . const . envHelper name
-                  . maybe EHNoProxy EHUseProxy
+proxyEnvironmentNamed name mp = ProxyOverride $ \secure' ->
+    systemProxyHelper (Just name) (httpProtocol secure') $ maybe EHNoProxy EHUseProxy mp
 
 -- | The default proxy settings for a manager. In particular: if the @http_proxy@ (or @https_proxy@) environment variable is set, use it. Otherwise, use the values in the @Request@.
 --
 -- Since 0.4.7
 defaultProxy :: ProxyOverride
-defaultProxy = ProxyOverride $ \secure ->
-    envHelper (envName secure) EHFromRequest
-
-data EnvHelper = EHFromRequest
-               | EHNoProxy
-               | EHUseProxy Proxy
-
-envHelper :: Text -> EnvHelper -> IO (Request -> Request)
-envHelper name eh = do
-    env <- getEnvironment
-    let lenv = Map.fromList $ map (first $ T.toLower . T.pack) env
-        lookupEnvVar n = lookup (T.unpack n) env <|> Map.lookup n lenv
-        noProxyDomains = domainSuffixes (lookupEnvVar "no_proxy")
-    case lookupEnvVar name of
-        Nothing  -> return noEnvProxy
-        Just ""  -> return noEnvProxy
-        Just str -> do
-            let invalid = throwIO $ InvalidProxyEnvironmentVariable name (T.pack str)
-            (p, muserpass) <- maybe invalid return $ do
-                uri <- case U.parseURI str of
-                    Just u | U.uriScheme u == "http:" -> return u
-                    _ -> U.parseURI $ "http://" ++ str
-
-                guard $ U.uriScheme uri == "http:"
-                guard $ null (U.uriPath uri) || U.uriPath uri == "/"
-                guard $ null $ U.uriQuery uri
-                guard $ null $ U.uriFragment uri
-
-                auth <- U.uriAuthority uri
-                port <-
-                    case U.uriPort auth of
-                        "" -> Just 80
-                        ':':rest ->
-                            case decimal $ T.pack rest of
-                                Right (p, "") -> Just p
-                                _ -> Nothing
-                        _ -> Nothing
-
-                Just $ (Proxy (S8.pack $ U.uriRegName auth) port, extractBasicAuthInfo uri)
-            return $ \req ->
-                if host req `hasDomainSuffixIn` noProxyDomains
-                then noEnvProxy req
-                else maybe id (uncurry applyBasicProxyAuth) muserpass
-                     req { proxy = Just p }
-    where noEnvProxy = case eh of
-            EHFromRequest -> id
-            EHNoProxy     -> \req -> req { proxy = Nothing }
-            EHUseProxy p  -> \req -> req { proxy = Just p  }
-          prefixed s | S8.head s == '.' = s
-                     | otherwise = S8.cons '.' s
-          domainSuffixes Nothing = []
-          domainSuffixes (Just "") = []
-          domainSuffixes (Just no_proxy) = [prefixed $ S8.dropWhile (== ' ') suffix | suffix <- S8.split ',' (S8.pack (map toLower no_proxy)), not (S8.null suffix)]
-          hasDomainSuffixIn host = any (`S8.isSuffixOf` prefixed (S8.map toLower host))
+defaultProxy = ProxyOverride $ \secure' ->
+    systemProxyHelper Nothing (httpProtocol secure') EHFromRequest
