@@ -8,16 +8,10 @@ module Network.HTTP.Client.Response
     , lbsResponse
     ) where
 
-import Control.Monad ((>=>), when)
-
-import Control.Exception (throwIO)
-
 import qualified Data.ByteString.Char8 as S8
 import qualified Data.ByteString.Lazy as L
 
-import Data.Default.Class (def)
-
-import Data.Maybe (isJust)
+import Data.Monoid (mempty)
 
 import qualified Network.HTTP.Types as W
 import Network.URI (parseURIReference, escapeURIString, isAllowedInURI)
@@ -122,6 +116,6 @@ getResponse connRelease timeout' req@(Request {..}) conn cont = do
         , responseVersion = version
         , responseHeaders = hs
         , responseBody = body
-        , responseCookieJar = def
+        , responseCookieJar = Data.Monoid.mempty
         , responseClose' = ResponseClose (cleanup False)
         }
